from ..ipaddress import IPAddress
from ..traffic import Traffic
from .utils import datetime_from_xmlrpc
from pyramid.view import view_config
from sqlalchemy.orm.exc import NoResultFound
import pyramid.security
import pyramid_xmlrpc


class RootFactory(object):
    """Traversal root factory to provide ACL for Ring 0."""

    __acl__ = [(pyramid.security.Allow, 'group:ring0', 'access')]

    def __init__(self, request):
        pass


@view_config(route_name='ring0', permission='access')
class Ring0(pyramid_xmlrpc.XMLRPCView):

    def check_consistency(self):
        """Return a list of errors about consistency within the directory."""
        # XXX might not be needed any longer
        return []

    def list_virtual_machines(self):
        # XXX convert to XML-RPC output format
        return []

    def list_nodes(self):
        return [self._format_node_for_api(n) for n in self.context.nodes]

    def lookup_node(self, lookup_name):
        for node in self.context.nodes:
            if node.name == lookup_name:
                return self._format_node_for_api(node)
        else:
            raise KeyError(lookup_name)

    def list_locations(self):
        return list(x.abbreviation for x in self.context.locations)

    def lookup_location(self, name):
        # XXX deprecated, use lookup_network instead
        for location in self.context.locations:
            if location == name:
                return location
        raise KeyError(name)

    def lookup_networks(self, location, af=''):
        details = self.lookup_networks_details(location, af)
        networks = {}
        for vlan, addrs in details.items():
            networks[vlan] = [n['cidr'] for n in addrs]
        return networks

    def lookup_networks_details(self, location, af=''):
        location = self.context.location_by_name(location)
        networks = {}
        if af:
            af = int(str(af).lower().replace('ipv', ''))
        for vlan in location.vlans:
            networks[vlan.class_.short] = addrs = []
            for network in vlan.networks:
                if af and network.netaddr.version != af:
                    continue
                addrs.append(dict(cidr=network.cidr,
                                  dhcp=network.dhcp_dynamic_range))
            addrs.sort(key=lambda x: x['cidr'])
            if not addrs:
                del networks[vlan.class_.short]
        return networks

    def list_networks(self):
        networks = set()
        for location in self.context.locations:
            for vlan in location.vlans:
                for network in vlan.networks:
                    networks.add(network.cidr)
        return list(sorted(networks))

    def lookup_interfaces(self, nodename, af_filter=''):
        node = self.context.node_by_name(nodename)
        interfaces = {}
        if af_filter:
            af_filter = int(af_filter.replace('ipv', ''))
        for nic in node.nics:
            addrs = []
            for addr in nic.addresses:
                if af_filter and addr.netaddr.version != af_filter:
                    continue
                addrs.append((addr.cidr, addr.reverse))
            interfaces[nic.vlan.class_.short] = sorted(addrs)
        return interfaces

    def lookup_addresses(self, nodename, vlan_filter='', af_filter=''):
        node = self.context.node_by_name(nodename)
        if af_filter:
            af_filter = int(af_filter.lower().replace('ipv', ''))
        addrs = set()
        for nic in node.nics:
            if vlan_filter and nic.vlan.class_.short != vlan_filter:
                continue
            for addr in nic.addresses:
                if af_filter and addr.netaddr.version != af_filter:
                    continue
                addrs.add(addr.cidr)
        return list(sorted(addrs))

    def list_resource_groups(self):
        return [rg.name for rg in self.context.resource_groups]

    def list_nodes_addresses(self, location_filter=None, vlan_filter=None,
                             af_filter=None):
        """Enumerate nodes with IP addresses, MAC addresses, RG, and ring.

        The list can optionally be filtered by VLAN, location, or address
        family.

        """
        addresses = filters.apply(
            self.context.addresses,
            VLAN=vlan_filter, Location=location_filter, AF=af_filter)
        return [dict(name=addr.resource.node.name,
                     ip=addr.cidr,
                     mac=addr.resource.mac,
                     vlan=addr.network.vlan.class_.short,
                     rg=addr.resource.node.resource_group.name,
                     ring=addr.resource.node.directory_ring)
                for addr in addresses]

    def list_redundant_nodes(self, location_filter=None, role_filter=None):
        groups = {}
        for group_id, nodes in self.context.list_redundant_nodes(
                location_filter, role_filter).items():
            groups[str(group_id)] = [node.name for node in nodes]
        return groups

    def schedule_maintenance(self, requests):
        node = zope.component.getUtility(
            gocept.directory.interfaces.INode,
            name=self.request.principal.id)
        for id, params in requests.items():
            request = gocept.directory.maintenance.Activity(
                datetime.timedelta(seconds=params['estimate']),
                params['comment'],
                id)
            node.register_activity(request)

        results = {}
        for activity in node.activities:
            if activity.ended:
                continue
            results[activity.reference_id] = result = {}
            result['time'] = (
                activity.starts.isoformat() if activity.starts else None)
        return results

    def end_maintenance(self, activities):
        node = zope.component.getUtility(
            gocept.directory.interfaces.INode,
            name=self.request.principal.id)
        for id, params in activities.items():
            activity = node.activity_by_reference_id(id)
            duration = params.get('duration')
            if duration is not None:
                duration = datetime.timedelta(seconds=params['duration'])
            activity.end(params['result'], duration)

    def store_traffic(self, timestamp, measured_values):
        timestamp = datetime_from_xmlrpc(timestamp)
        for ip, traffic in measured_values:
            if not traffic:
                continue
            try:
                ip = IPAddress.query().filter(IPAddress.address == ip).one()
            except NoResultFound:
                continue
            Traffic.store(timestamp, ip, traffic)
        return ''       # XXX get allow_none implemented
