from sqlalchemy import Column, Integer
import alembic.config
import alembic.migration
import alembic.script
import sqlalchemy
import sqlalchemy.exc
import sqlalchemy.ext.declarative
import sqlalchemy.orm
import sqlalchemy.types
import transaction
import zope.component
import zope.event
import zope.interface
import zope.sqlalchemy


class IDatabase(zope.interface.Interface):
    pass


class AlembicContext(object):

    def __init__(self, conn):
        config = alembic.config.Config()
        config.set_main_option("script_location", "gocept.directory2:alembic")
        self.migration_context = (
            alembic.migration.MigrationContext.configure(conn))
        self.script = alembic.script.ScriptDirectory.from_config(config)


class alembic_context(object):
    def __init__(self, engine):
        self.engine = engine

    def __enter__(self):
        self.conn = conn = self.engine.connect()
        return AlembicContext(conn)

    def __exit__(self, exc_type, exc_value, tb):
        self.conn.close()
        if exc_type is not None:
            raise exc_type(exc_value)


class Database(object):

    zope.interface.implements(IDatabase)

    def __init__(self, dsn, engine_args={}, testing=False):
        self.dsn = dsn
        self.testing = testing
        self._verify()
        self.engine = sqlalchemy.create_engine(dsn, **engine_args)
        self.session_factory = sqlalchemy.orm.scoped_session(
            sqlalchemy.orm.sessionmaker(
                bind=self.engine,
                extension=zope.sqlalchemy.ZopeTransactionExtension(
                    keep_session=testing)))

    def _verify(self):
        # Step 1: Try to identify a testing table
        engine = sqlalchemy.create_engine(self.dsn)
        conn = engine.connect()
        try:
            conn.execute("SELECT * FROM tmp_functest")
        except sqlalchemy.exc.ProgrammingError:
            db_is_testing = False
        else:
            db_is_testing = True
        conn.invalidate()

        if (self.testing and db_is_testing):
            # We're running as a test and it is a test dabase. Continue
            return
        if (not self.testing and not db_is_testing):
            # We're running as a production system and we have a production
            # database. Continue.
            return

        # We're not in a valid state. Bail out.
        raise SystemExit("Not working against correct database (live vs "
                         "testing). Refusing to set up database connection.")

    def assert_database_revision_is_current(self):
        with alembic_context(self.engine) as ac:
            head = ac.script.get_current_head()
            db_rev = ac.migration_context.get_current_revision()
            if head != db_rev:
                raise ValueError("Database revision {} does not match current "
                                 "revision {}.\nMaybe you want to call "
                                 "`bin/alembic upgrade head`.".format(
                                     db_rev, head))

    def initialize_database(self):
        from .user import HumanUser
        from .vlan import VLANClass
        from .resourcegroup import ResourceGroup
        try:
            with transaction.manager:
                VLANClass.query().all()
        except sqlalchemy.exc.ProgrammingError:
            pass
        else:
            raise ValueError("Database already exists. "
                             "Refusing to initialize again.")
        Object.metadata.create_all(self.engine)
        VLANClass.create_default_vlan_classes()
        ResourceGroup.create_default_resource_groups()
        HumanUser.create(uid='admin', firstname='Ad', lastname='Min',
                         is_admin=True, password='admin')
        #mark the database to be in the latest revision:
        with alembic_context(self.engine) as ac:
            ac.migration_context._update_current_rev(
                ac.migration_context.get_current_revision(),
                ac.script.get_current_head())

    @property
    def session(self):
        return self.session_factory()

    def add(self, obj):
        return self.session.add(obj)

    def delete(self, obj):
        return self.session.delete(obj)

    def query(self, *args, **kw):
        return self.session.query(*args, **kw)


class Added(zope.interface.interfaces.ObjectEvent):
    """An object has been created and added to the session."""

    def __init__(self, object):
        self.object = object


class ObjectBase(object):

    @staticmethod
    def db():
        return zope.component.getUtility(IDatabase)

    @classmethod
    def create(cls, **kw):
        """Canonical way to create an object of this class."""
        obj = cls()
        for key, value in kw.items():
            if key not in dir(cls):
                raise KeyError('%r is not known on the class. Typo?' % key)
            setattr(obj, key, value)
        obj.persist()
        return obj

    def persist(self):
        """Add make the newly created object known to the database."""
        self.db().add(self)
        zope.event.notify(Added(self))

    def delete(self):
        """Remove the object from the database."""
        self.db().delete(self)

    @classmethod
    def query(cls):
        return cls.db().query(cls)

    @classmethod
    def find_by_id(cls, id):
        return cls.db().query(cls).filter_by(id=id).first()

    @classmethod
    def context_factory(cls, request):
        return cls.query().filter_by(**request.matchdict).first()


class ObjectWithId(ObjectBase):

    id = Column(Integer, primary_key=True)


Object = sqlalchemy.ext.declarative.declarative_base(cls=ObjectWithId)


class ListOfStrings(sqlalchemy.types.TypeDecorator):
    '''Turns lists into newline separated strings.

    Ensures that the list items do not contain newlines.

    '''

    impl = sqlalchemy.types.Text

    def process_bind_param(self, value, dialect):
        if value is None:
            return
        if '\n' in ''.join(value):
            raise ValueError(
                'List of strings may not contain '
                'newline characters: %r' % value)
        return '\n'.join(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return
        return value.split('\n')

    # XXX the sample showed this, but I don't know what it does.
    #def copy(self):
    #    return MyType(self.impl.length)
