from .maintenance import Planning
from .numberrange import NumberRange, NumberRangeFactory
from gocept.directory2.db import ListOfStrings
from sqlalchemy import Column, Integer, Text, Date, ForeignKey, String
from sqlalchemy import Table, Boolean
import gocept.directory2.db
import grokcore.component as grok
import sqlalchemy.orm


class Node(gocept.directory2.db.Object, Planning):
    """Physical or virtual machine."""

    __tablename__ = 'nodes'
    class_ = Column(String(50))
    __mapper_args__ = {'polymorphic_on': class_}

    resource_group_id = Column(
        Integer, ForeignKey('resourcegroups.id'), nullable=False)
    resource_group = sqlalchemy.orm.relation(
        'ResourceGroup', uselist=False,
        backref=sqlalchemy.orm.backref('nodes'))

    name = Column(Text, unique=True, nullable=False)
    global_id = Column(Integer, unique=True)

    location_id = Column(Integer, ForeignKey('locations.id'), nullable=False)
    location = sqlalchemy.orm.relation(
        'Location', backref=sqlalchemy.orm.backref('nodes'))

    service_description = Column(String)
    directory_password = Column(String)  # BytesLine
    directory_ring = Column(Integer)
    redundancy_group = Column(Integer)

    puppet_classes = Column(ListOfStrings, default=[])
    environment = Column(String)

    # Planning
    auto_schedule = Column(Boolean, default=True)

    default_nics = []
    generate_default_mac = False

    def delete(self):
        # XXX SQLAlchemy delete() workaround.
        for nic in self.nics.values():
            nic.delete()
        super(Node, self).delete()

    def check_password(self, value):
        return value == self.directory_password


@grok.subscribe(Node, gocept.directory2.db.Added)
def set_node_global_id(node, event):
    """Set the id number at resource group add."""
    node.global_id = NodeGlobalIDRange().next()


@NumberRangeFactory
class NodeGlobalIDRange(NumberRange):

    type = 'node_global_id'
    min = 4100
    max = 65535

    def is_available(self, number):
        return not Node.query().filter_by(global_id=number).count()


physicalmachine_rackunits_table = Table(
    'physicalmachines_rackunits',
    gocept.directory2.db.Object.metadata,
    Column('physicalmachine_id', Integer, ForeignKey('physicalmachines.id')),
    Column('rackunit_id', Integer, ForeignKey('rackunits.id')))


class PhysicalMachine(Node):

    __tablename__ = 'physicalmachines'
    id = Column(
        'id', Integer, ForeignKey('nodes.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'physical'}

    kvm_memory = Column(Integer)
    kvm_disk = Column(Integer)
    end_of_support_contract = Column(Date)

    rackunits = sqlalchemy.orm.relation(
        'RackUnit', secondary=physicalmachine_rackunits_table,
        backref='machines')

    default_nics = ['srv', 'fe', 'sto', 'mgm', 'ipmi']
    generate_default_mac = False


class VirtualMachine(Node):

    __tablename__ = 'virtualmachines'
    id = Column(
        'id', Integer, ForeignKey('nodes.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'virtual'}

    memory = Column(Integer)  # min=256, not bool(x % 128)
    disk = Column(Integer)  # min=5, not bool(x % 5)
    cores = Column(Integer)  # 1-12, default=1

    kvm_host_id = Column(Integer, ForeignKey('physicalmachines.id'))
    kvm_host = sqlalchemy.orm.relation(
        'PhysicalMachine', uselist=False, backref='virtual_machines',
        primaryjoin=PhysicalMachine.id == kvm_host_id)

    default_nics = ['srv', 'fe']
    generate_default_mac = True
