from .location import Location
from .resourcegroup import ResourceGroup
from .vlan import VLANClass
from wsgiref.simple_server import WSGIRequestHandler
import gocept.directory2.application
import gocept.directory2.db
import gocept.testdb
import plone.testing
import plone.testing.zca
import sqlalchemy
import sqlalchemy.engine.reflection
import transaction
import zope.testbrowser.wsgi
import tempfile
import os


class ApplicationLayer(plone.testing.Layer):

    defaultBases = [plone.testing.zca.LAYER_CLEANUP]

    def __init__(self, factory):
        super(ApplicationLayer, self).__init__()
        self.factory = factory

    def setUp(self):
        plone.testing.zca.pushGlobalRegistry()
        self['application'] = self.factory()
        self['application'].configure()
        self._setup_test_runtime()

    def _setup_postgres_db(self):
        self['db'] = db = gocept.testdb.PostgreSQL()
        if db.exists:
            raise ValueError("Database {}@{} already exists!".format(
                db.db_name, db.db_host))
        db.create()
        database = gocept.directory2.db.Database(db.dsn, testing=True)
        gocept.directory2.db.Object.metadata.create_all(database.engine)
        zope.component.provideUtility(database)

    def _setup_domainshuttle_db(self):
        self['db_domainshuttle'] = db_domainshuttle = gocept.testdb.MySQL(
            cmd_postfix=os.environ.get('MYSQL_CMD_POSTFIX', ''))
        db_domainshuttle.create()
        database_domainshuttle = gocept.directory2.db_domainshuttle.Database(
            db_domainshuttle.dsn, testing=True)
        gocept.directory2.db_domainshuttle.Object.metadata.create_all(
            database_domainshuttle.engine)
        zope.component.provideUtility(database_domainshuttle)

    def _setup_test_runtime(self):
        # XXX this is the converse of the setup_runtime method.
        self['application'].config.include('pyramid_mailer.testing')
        self._setup_postgres_db()
        self._setup_domainshuttle_db()
        notifier = gocept.directory2.maintenance.Notifier()
        zope.component.provideUtility(notifier)

    def _teardown_postgres(self):
        self['db'].drop()
        del self['db']

    def _teardown_domainshuttle(self):
        self['db_domainshuttle'].drop()
        del self['db_domainshuttle']

    def tearDown(self):
        del self['application']
        self.close_all_connections()
        self._teardown_postgres()
        self._teardown_domainshuttle()
        plone.testing.zca.popGlobalRegistry()

    def _close_postgres(self):
        db = zope.component.getUtility(gocept.directory2.db.IDatabase)
        db.engine.dispose()

    def _close_domainshuttle(self):
        db = zope.component.getUtility(gocept.directory2.db_domainshuttle.IDatabase)
        db.engine.dispose()

    def close_all_connections(self):
        # close all connections, but...
        transaction.abort()
        # ...sometimes transaction.abort() is not enough, and...
        self._close_postgres()
        self._close_domainshuttle()
        # ...connections that have been checked-out from the pool and not yet
        # returned are not closed by dispose, either, so we have to hunt them
        # down ourselves:
        for conn in sqlalchemy.pool._refs:
            conn.close()

    def testSetUp(self):
        transaction.abort()
        self.truncate_all_tables_postgres()
        self.truncate_all_tables_domainshuttle()
        self.create_fixture()
        # Commit and prepare a new session before the test starts.
        transaction.commit()
        self.db.session.close_all()
        self.db_domainshuttle.session.close_all()

    def testTearDown(self):
        del self['rg0']
        del self['loc0']
        self.db.session.close_all()
        self.db_domainshuttle.session.close_all()

    def truncate_all_tables_postgres(self):
        tables = ', '.join('"%s"' % x for x in self.table_names(self.db))
        self.db.session.execute(
            'TRUNCATE %s RESTART IDENTITY' % tables)
        zope.sqlalchemy.mark_changed(self.db.session)
        transaction.commit()

    def truncate_all_tables_domainshuttle(self):
        session = self.db_domainshuttle.session
        session.execute('SET FOREIGN_KEY_CHECKS=0;')
        for table_name in self.table_names(self.db_domainshuttle):
            session.execute('TRUNCATE %s;' % table_name)
        session.execute('SET FOREIGN_KEY_CHECKS=1;')
        zope.sqlalchemy.mark_changed(self.db_domainshuttle.session)
        transaction.commit()

    def table_names(self, db):
        inspector = sqlalchemy.engine.reflection.Inspector.from_engine(
            db.session.bind)
        return inspector.get_table_names()

    def create_fixture(self):
        VLANClass.create_default_vlan_classes()
        ResourceGroup.create_default_resource_groups()
        self['rg0'] = ResourceGroup.create(name='rg0')
        self['loc0'] = Location.create(name='loc0')

    @property
    def db(self):
        return zope.component.getUtility(gocept.directory2.db.IDatabase)

    @property
    def db_domainshuttle(self):
        return zope.component.getUtility(
            gocept.directory2.db_domainshuttle.IDatabase)

# This layer is used for a basic application setup without any external
# interface settings like API, Admin UI, ... This is the pure domain model
# environment.
BASE_APPLICATION_LAYER = ApplicationLayer(
    gocept.directory2.application.ApplicationFactory)


class WSGILayer(plone.testing.Layer):
    """WSGILayer - requires an ApplicationLayer as base layer."""

    def setUp(self):
        self['application'].settings = {'beaker.lock_dir': '/tmp'}
        self['wsgi_app'] = self['application'].make_wsgi_app()

    def tearDown(self):
        del self['wsgi_app']


class LogWSGIRequestHandler(WSGIRequestHandler):

    log = False

    def log_request(self, *args):
        if self.log:
            WSGIRequestHandler.log_request(self, *args)


class BrowserLayer(plone.testing.Layer):

    def testSetUp(self):
        self['browser'] = TestBrowser(wsgi_app=self['wsgi_app'])

    def testTearDown(self):
        del self['browser']


class TestBrowser(zope.testbrowser.wsgi.Browser):

    def login_basic_auth(self, username, password):
        self.addHeader('Authorization', zope.testbrowser.wsgi.auth_header(
            'Basic %s:%s' % (username, password)))

    @property
    def logged_in(self):
        return 'Authorization' in dict(self.mech_browser.addheaders)

    def open(self, *args, **kw):
        transaction.commit()
        return super(TestBrowser, self).open(*args, **kw)
