from ..permission import Permission
from ..resourcegroup import ResourceGroup, ResourceGroupNumberRange
from ..user import HumanUser, ServiceUser
from ..machine import Node
import unittest
import gocept.directory2.testing


class ResourceGroupUTests(unittest.TestCase):
    """Unit tests for ..resourcegroup.ResourceGroup."""

    def _makeOne(self):
        return ResourceGroup()

    def test_technical_contacts_returns_empty_list_for_None(self):
        rg = self._makeOne()
        rg._technical_contacts = None
        self.assertEqual([], rg.technical_contacts)

    def test_technical_contacts_returns_data_as_list(self):
        rg = self._makeOne()
        rg._technical_contacts = 'a\nb'
        self.assertEqual(['a', 'b'], rg.technical_contacts)

    def test_technical_contacts_stores_list_as_plain_data(self):
        rg = self._makeOne()
        rg.technical_contacts = ['a', 'b']
        self.assertEqual('a\nb', rg._technical_contacts)

    def test_technical_contacts_strips_values_on_store(self):
        rg = self._makeOne()
        rg.technical_contacts = ['a ', ' b']
        self.assertEqual('a\nb', rg._technical_contacts)

    def test_resourcegroup_maintenance_duration(self):
        rg = ResourceGroup()
        rg.maintenance_start = 1
        rg.maintenance_end = 2
        self.assertEquals(1, rg.maintenance_duration)
        rg.maintenance_start = 2
        rg.maintenance_end = 2
        self.assertEquals(24, rg.maintenance_duration)
        rg.maintenance_start = 22
        rg.maintenance_end = 4
        self.assertEquals(6, rg.maintenance_duration)


class ResourceGroupFTests(unittest.TestCase):
    """Functional tests for ..resourcegroup.ResourceGroup."""

    layer = gocept.directory2.testing.BASE_APPLICATION_LAYER

    def test_id_number_is_generated_at_creation(self):
        rg1 = ResourceGroup.create()
        rg2 = ResourceGroup.create()
        self.assertNotEqual(rg1.id_number, rg2.id_number)

    def test_suggest_on_unnamed_resourcegroup(self):
        rg = ResourceGroup.create(name='')
        self.assertEquals(rg.suggest_hostname(), '')

    def test_suggest_first(self):
        rg = ResourceGroup.create(name='rg')
        self.assertEquals(rg.suggest_hostname(), 'rg00')

    def test_suggest_second(self):
        rg = ResourceGroup.create(name='rg')
        Node.create(name='rg00',
                    resource_group=rg,
                    location=self.layer['loc0'])
        self.assertEquals(rg.suggest_hostname(), 'rg01')

    def test_suggest_onehundretandone(self):
        rg = ResourceGroup.create(name='rg')
        for i in range(100):
            nodename = 'rg%02d' % i
            Node.create(name=nodename, resource_group=rg,
                        location=self.layer['loc0'])
        self.assertEquals(rg.suggest_hostname(), 'rg100')

    def test_duplicate_group_name_causes_validation_error(self):
        ResourceGroup.create(name='rg')
        self.assertRaises(Exception, ResourceGroup.create, name='rg')

    def test_name_collision_groups_permissions_causes_validation_error(self):
        Permission.create(name='login')
        self.assertRaises(Exception, ResourceGroup.create(name='login'))

    def test_legacy_permission_numbers_are_not_assigned_to_groups(self):
        Permission.create(id_number=2046)
        range = ResourceGroupNumberRange()
        range.next_number = 2045
        self.assertEqual(2045, range.next())
        self.assertEqual(2047, range.next())
        self.assertEqual(2048, range.next())


class ResourceGroupAssociatedWithHumanUserTests(unittest.TestCase):

    layer = gocept.directory2.testing.BASE_APPLICATION_LAYER

    def setUp(self):
        super(ResourceGroupAssociatedWithHumanUserTests, self).setUp()
        self.rg = ResourceGroup.query().filter(
            ResourceGroup.name == 'rg0').one()
        self.user = self._create_user('hans')

    def _create_user(self, uid):
        return HumanUser.create(uid=uid, firstname=uid, lastname=uid)

    def test_adding_user_to_resource_group_should_work_both_ways(self):
        self.assertEqual([], list(self.user.resource_groups))
        self.assertEqual([], list(self.rg.users))
        self.rg.add_user(self.user)
        self.assertEqual([self.rg], list(self.user.resource_groups))
        self.assertEqual([self.user], list(self.rg.users))

    def test_adding_users_to_resource_group_adds_each_only_once(self):
        user2 = self._create_user('klaus')
        self.rg.add_user(self.user)
        self.rg.add_user(user2)
        self.rg.add_user(self.user)
        self.assertEqual(1, len(self.user.resource_groups))
        self.assertEqual(2, len(self.rg.users))

    def test_removing_user_from_resource_group_should_work_both_ways(self):
        self.rg.add_user(self.user)
        self.rg.remove_user(self.user)
        self.assertEqual([], list(self.user.resource_groups))
        self.assertEqual([], list(self.rg.users))

    def test_removing_user_from_resource_group_should_work_if_unassoc(self):
        self.rg.remove_user(self.user)
        self.assertEqual([], list(self.user.resource_groups))
        self.assertEqual([], list(self.rg.users))

    def test_deleting_resource_group_should_disassociate_from_user(self):
        self.rg.add_user(self.user)
        self.user.delete()
        self.assertEqual([], list(self.user.resource_groups))
        self.assertEqual([], list(self.rg.users))


class ResourceGroupAssociatedWithServiceUserTests(
        ResourceGroupAssociatedWithHumanUserTests):

    def _create_user(self, uid):
        return ServiceUser.create(uid=uid)


class ResourceGroupNumberRangeTests(unittest.TestCase):

    layer = gocept.directory2.testing.BASE_APPLICATION_LAYER

    def setUp(self):
        super(ResourceGroupNumberRangeTests, self).setUp()
        rg = ResourceGroup.create()
        rg.id_number = 12345
        Permission.create(id_number=2123)

    def test_is_available_returns_false_for_existing_resource_group(self):
        self.assertFalse(ResourceGroupNumberRange().is_available(12345))

    def test_is_available_returns_true_for_not_existing_resource_group(self):
        self.assertTrue(ResourceGroupNumberRange().is_available(12344))
        self.assertTrue(ResourceGroupNumberRange().is_available(12346))

    def test_is_available_returns_false_for_existing_permission_id(self):
        self.assertFalse(ResourceGroupNumberRange().is_available(2123))
