# coding: utf-8
from gocept.directory2.db import ListOfStrings
from passlib.hash import sha512_crypt
from sqlalchemy import Column, ForeignKey, Table
from sqlalchemy import Integer, Text, String, Boolean
from sqlalchemy.orm import relation, backref
import gocept.directory2.db
import grokcore.component as grok
import os.path
import passogva
import zope.interface

CRYPT_PREFIX = '{CRYPT}'
PASSWORD_LENGTH = 12

PASSWORD_MAIL_CONTENT = dict(
    de=u"""\
für Ihr Konto bei gocept.net wurde gerade ein neues Passwort vergeben. Es
lautet:

        %(password)s

(gesprochen: "%(hyphenated)s").
""",
    en=u"""\
a new password has been set for your gocept.net account. It reads:

        %(password)s

(pronounced: "%(hyphenated)s").
""",
)


PASSWORD_MAIL_SUBJECT = dict(
    de=u'Ihr neues Passwort für gocept.net',
    en=u'Your new password for gocept.net',
)

users_resourcegroups = Table(
    'users_resourcegroups',
    gocept.directory2.db.Object.metadata,
    Column('user_id', Integer, ForeignKey('users.id')),
    Column('resourcegroup_id', Integer, ForeignKey('resourcegroups.id')))


class User(gocept.directory2.db.Object):
    """Human or service user."""

    __tablename__ = 'users'
    class_ = Column(String(50))
    __mapper_args__ = {'polymorphic_on': class_}

    id_number = Column(Integer, unique=True)
    gid_number = Column(Integer)

    uid = Column(String, unique=True, nullable=False)
    name = Column(Text)     # XXX Name for service users? Vs. UID?
    description = Column(Text)

    encrypted_password = Column(String)

    mail = Column(String)
    login_shell = Column(String)
    home_directory = Column(String)
    ssh_pubkey = Column(ListOfStrings, default=[])

    resource_groups = relation(
        'ResourceGroup', collection_class=set, secondary=users_resourcegroups,
        cascade='all',
        backref=backref('users', collection_class=set))

    home_dir_base = '/home'

    def __init__(self, uid):
        self.home_directory = os.path.join(self.home_dirs, self.uid)

    def __get_password(self):
        return self.encrypted_password

    def __set_password(self, value):
        if value:
            value = CRYPT_PREFIX + sha512_crypt.encrypt(value)
        self.encrypted_password = value

    password = property(__get_password, __set_password)

    def delete(self):
        # XXX SQLAlchemy delete() workaround.
        self.resource_groups = set()

    @property
    def password_set(self):
        return bool(self.password)

    def check_password(self, given_password):
        if self.password is None:
            return False
        if not self.password.startswith(CRYPT_PREFIX):
            return False
        password = self.password.replace(CRYPT_PREFIX, '')
        return sha512_crypt.verify(given_password, password)

    @staticmethod
    def generate_password():
        password, hyphenated = passogva.generate_password(
            minlen=PASSWORD_LENGTH, maxlen=PASSWORD_LENGTH)
        return unicode(password), hyphenated

    def reset_password(self):
        password, hyphenated = self.generate_password()
        self.password = password
        grok.notify(PasswordGeneratedEvent(self, password, hyphenated))

    def mail_password(self, clear_password, hyphenated_password):
        if not self.mail:
            return
        lang = 'en'  # XXX make this a property of the user
        content = PASSWORD_MAIL_CONTENT[lang] % dict(
            password=clear_password, hyphenated=hyphenated_password)
        subject = PASSWORD_MAIL_SUBJECT[lang]
        recipients = [self.mail]
        log_origin = self.__name__
        # XXX sendmail
        print(log_origin, recipients, lang, subject, content)


class HumanUser(User):

    __tablename__ = 'users_human'
    id = Column(
        'id', Integer, ForeignKey('users.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'human'}

    firstname = Column(Text, nullable=False)
    lastname = Column(Text, nullable=False)
    pager = Column(String)
    is_admin = Column(Boolean, nullable=False, default=False)

    def __init__(self):
        self.gid_number = 100

    @property
    def name(self):
        return ('%s %s' % (self.firstname, self.lastname)
                if self.firstname and self.lastname else self.lastname)


class ServiceUser(User):

    __tablename__ = 'users_service'
    id = Column(
        'id', Integer, ForeignKey('users.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'service'}

    home_dir_base = '/srv'

    def __init__(self):
        self.gid_number = 101


class IPasswordGeneratedEvent(zope.interface.Interface):

    user = zope.interface.Attribute('User')
    password = zope.interface.Attribute('Clear-text password')
    hyphenated = zope.interface.Attribute('Pronounciation hint for password')


class PasswordGeneratedEvent(object):

    grok.implements(IPasswordGeneratedEvent)

    def __init__(self, user, password, hyphenated):
        self.user = user
        self.password = password
        self.hyphenated = hyphenated
